@@ -8,6 +8,8 @@ from tornado.options import define, options
 import base64
 from tornado.httputil import parse_multipart_form_data
 from tornado.ioloop import IOLoop
+from tornado.gen import coroutine, Return, engine
+from tornado.concurrent import return_future, Future
 import logging
 
 define("allow_origin", default="*", help="This is the value for the Access-Control-Allow-Origin header (default *)")
@@ -66,7 +68,8 @@ class TotoHandler(RequestHandler):
   The request handler will be passed as the first parameter to the invoke function and
   provides access to the server's database connection, the current session and other
   useful request properties. Request parameters will be passed as the second argument
-  to the invoke function.
+  to the invoke function. Any return values from ``invoke()`` functions should be
+  JSON serializable.
   
   Toto methods are generally invoked via a POST request to the server with a JSON
   serialized object as the body. The body should contain two properties:
@@ -80,6 +83,13 @@ class TotoHandler(RequestHandler):
 
   Will call method_module.account.create.invoke(handler, {'user_id': 'test', 'password': 'testpassword'})
 
+  An ``invoke()`` function can be decorated with ``@tornado.gen.coroutine`` and be run as a Tornado coroutine.
+
+  Alternatively, an ``invoke(handler, parameters)`` function may be decorated with ``@toto.invocation.asynchronous``.
+  If a function decorated in this manner does not return a value, the connection well remain open until
+  ``handler.respond(result, error)`` is called where ``error`` is an ``Exception`` or ``None`` and ``result``
+  is a normal ``invoke()`` return value or ``None``.
+
   There are client libraries for iOS and Javascript that will make using Toto much easier. They are
   available at https://github.com/JeremyOT/TotoClient-iOS and https://github.com/JeremyOT/TotoClient-JS
   respectively.
@@ -188,6 +198,7 @@ class TotoHandler(RequestHandler):
     logging.error("TotoException: %s Value: %s" % (e.code, e.value))
     return e.__dict__
 
+  @coroutine
   def invoke_method(self, path, request_body, parameters, handler=None):
     result = None
     error = None
@@ -195,10 +206,15 @@ class TotoHandler(RequestHandler):
     try:
       method = self.__get_method(self.__get_method_path(path, request_body))
       self.__active_methods.append(method)
-      result = method.invoke(handler or self, parameters)
+      output = method.invoke(handler or self, parameters)
+      if isinstance(output, Future):
+        #result is a future, so yield the real response
+        result = yield output
+      else:
+        result = output
     except Exception as e:
       error = self.error_info(e)
-    return result, error, (hasattr(method, 'asynchronous'))
+    raise Return((result, error, (hasattr(method, 'asynchronous'))))
 
   def options(self, path=None):
     allowed_headers = set(['x-toto-hmac','x-toto-session-id','origin','content-type'])
@@ -211,12 +227,12 @@ class TotoHandler(RequestHandler):
     self.add_header('access-control-allow-methods', ','.join(self.SUPPORTED_METHODS))
     self.add_header('access-control-expose-headers', 'x-toto-hmac')
   
-  @tornado.web.asynchronous
+  @coroutine
   def head(self, path=None):
     self.headers_only = True
     self.get(path)
 
-  @tornado.web.asynchronous
+  @coroutine
   def get(self, path=None):
     parameters = {}
     # Convert parameters with one item to string, will cause undesired behavior if user means to pass array with length 1
@@ -225,9 +241,9 @@ class TotoHandler(RequestHandler):
         parameters[k] = v[0]
       else:
         parameters[k] = v
-    self.process_request(path, self.body, parameters)
+    yield self.process_request(path, self.body, parameters)
 
-  @tornado.web.asynchronous
+  @coroutine
   def post(self, path=None):
     content_type = 'content-type' in self.request.headers and self.request.headers['content-type'] or 'application/json'
     if not content_type.startswith('application/json'):
@@ -244,11 +260,14 @@ class TotoHandler(RequestHandler):
     else:
       self.body = json.loads(self.request.body)
     if self.body and 'batch' in self.body:
-      self.batch_process_request(self.body['batch'])
+      yield self.batch_process_request(self.body['batch'])
     else:
-      self.process_request(path, self.body, self.body and 'parameters' in self.body and self.body['parameters'] or {})
+      yield self.process_request(path, self.body, self.body and 'parameters' in self.body and self.body['parameters'] or {})
   
-  def batch_process_request(self, requests):
+  @return_future
+  @engine
+  def batch_process_request(self, requests, callback):
+    self._request_callback = callback
     self.session = None
     self.add_header('access-control-allow-origin', self.ACCESS_CONTROL_ALLOW_ORIGIN)
     self.add_header('access-control-expose-headers', 'x-toto-hmac')
@@ -256,23 +275,26 @@ class TotoHandler(RequestHandler):
     self.batch_results = {}
     for k, v in ((i, requests[i]) for i in self.request_keys):
       proxy = BatchHandlerProxy(self, k)
-      (result, error, async) = self.invoke_method(None, v, v.get('parameters', {}), handler=proxy)
+      result, error, async = yield self.invoke_method(None, v, v.get('parameters', {}), handler=proxy)
       if async:
         proxy.async = True
       if result or error or not async:
         proxy.respond(result, error, allow_async=False)
 
-  def process_request(self, path, request_body, parameters):
+  @return_future
+  @engine 
+  def process_request(self, path, request_body, parameters, callback):
+    self._request_callback = callback
     self.session = None
     self.add_header('access-control-allow-origin', self.ACCESS_CONTROL_ALLOW_ORIGIN)
     self.add_header('access-control-expose-headers', 'x-toto-hmac')
-    (result, error, async) = self.invoke_method(path, request_body, parameters)
+    result, error, async = yield self.invoke_method(path, request_body, parameters)
     if async:
       self.async = True
     if result is not None or error:
       self.respond(result, error, allow_async=False)
     elif not async and not self._finished:
-      self.finish()
+      self._request_callback()
 
   def respond(self, result=None, error=None, batch_results=None, allow_async=True):
     '''Respond to the request with the given result or error object (the ``batch_results`` and
@@ -329,7 +351,7 @@ class TotoHandler(RequestHandler):
     if not self.headers_only:
       self.write(body)
     if finish:
-      self.finish()
+      self._request_callback()
 
   def on_connection_close(self):
     '''You should not call this method directly, but if you implement an ``on_connection_close()`` function in a
